import param
import panel as pn
import httpx
import json
import eve
from .eve_model import EveModelBase
from .auth import EveAuthBase, AuthSelector
from . import settings


APPS = {}

class EveHttpClient(EveModelBase):
    pass

class EveHttpxClient(EveHttpClient):
    _app_settings = param.Dict(default=None, allow_None=True)
    _self_serve = param.Boolean(False)
    server_url = param.String(default="http://localhost:5000", regex=r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))")
    server_urls = param.List(default=[], class_=str)
    auth = param.ClassSelector(EveAuthBase, default=AuthSelector())
    _log = param.String()
    _messages = param.List(default=[])
    _busy = param.Boolean(False)

    def __init__(self, **params):
        super().__init__(**params)

    @classmethod
    def from_app_settings(cls, settings, urls=None):
        if urls is None:
            urls = settings.get("SERVERS", ["http://localhost:5000"])
        urls = [url.strip("/") + "/".join([settings.get("URL_PREFIX",""), settings.get("API_VERSION", "")]).replace("//","/").replace("///","/") for url in urls]
        instance = cls(server_url=urls[0], server_urls=urls, _app_settings=settings)
        return instance
    
    @property
    def app(self):
        if not self._self_serve:
            return None
        if self.name not in APPS and self._app_settings:
            APPS[self.name] = eve.Eve(self._app_settings)
        return APPS.get(self.name, None)

    def headers(self):
        headers = self.auth.get_headers()
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"
        return headers
    
    def get(self, url, timeout=10, **params):
        with httpx.Client(app=self.app, base_url=self.server_url) as client:
            self._busy = True
            try:
                resp = client.get(url, params=params, headers=self.headers(), timeout=timeout)
                self._busy = False
                if resp.is_error:
                    self.log_error(resp.text)
                else:
                    self.clear_messages()
                    return resp.json()
            except Exception as e:
                self.log_error(e)
            self._busy = False
        return []

    def post(self, url, data="", json={}, timeout=10):
        with httpx.Client(app=self.app, base_url=self.server_url) as client:
            self._busy = True
            try:
                resp = client.post(url, data=data, json=json, headers=self.headers(), timeout=timeout)
                self._busy = False
                if resp.is_error:
                    self.log_error(resp.text)
                    return False
                else:
                    self.clear_messages()
                    return True
            except Exception as e:
                self.log_error(e)
            self._busy = False

    def put(self, url, data, timeout=10):
        with httpx.Client(app=self.app, base_url=self.server_url) as client:
            self._busy = True
            try:
                resp = client.put(url, data=data, headers=self.headers(), timeout=timeout)
                self._busy = False
                if resp.is_error:
                    self.log_error(resp.text)
                    return False
                else:
                    self.clear_messages()
                    return True
            except Exception as e:
                self.log_error(e)
        self._busy = False

    def patch(self, url, data, timeout=10):
        with httpx.Client(app=self.app, base_url=self.server_url) as client:
            self._busy = True
            try:
                resp = client.patch(url, data=data, headers=self.headers(), timeout=timeout)
                self._busy = False
                if resp.is_error:
                    self.log_error(resp.text)
                    return False
                else:
                    self.clear_messages()
                    return True
            except Exception as e:
                self.log_error(e)
            self._busy = False

    def delete(self, url, etag, timeout=10):
        with httpx.Client(app=self.app, base_url=self.server_url) as client:
            self._busy = True
            headers = self.headers()
            if etag:
                headers["If-Match"] = etag
            try:
                resp = client.post(url,headers=headers, timeout=timeout)
                self._busy = False
                if resp.is_error:
                    self.log_error(resp.text)
                    return False
                else:
                    self.clear_messages()
                    return True
            except Exception as e:
                self.log_error(e)
            self._busy = False

    def log_error(self, e):
        self._log = str(e)
        self._messages = [str(e)]

    def clear_messages(self):
        self._messages = []
    
    def set_token(self, token):
        self.auth.set_token(token)

    @param.depends("_messages")
    def messages(self):
        messages = [pn.pane.Alert(msg, alert_type="danger", width=settings.GUI_WIDTH-20) for msg in self._messages]
        return pn.Column(*messages, width=settings.GUI_WIDTH)

    def make_panel(self):
        settings = pn.Param(self.param, parameters=["_self_serve", "server_url","_log"],
                        width=500, height=150, show_name=False,
                        widgets={"_log": {'type': pn.widgets.TextAreaInput, 'disabled': True, "height":150},
                                "server_url": {'type': pn.widgets.AutocompleteInput, "options": self.server_urls },
                                }
                                )
        return pn.Column(self.auth.panel, settings)
    
    @param.depends("_busy")
    def busy_indicator(self):
        return pn.Param(self.param._busy, width=25,
                    widgets={"_busy": {"type": pn.indicators.LoadingSpinner,"color":"primary", "width":20, "height":20}})

DEFAULT_HTTP_CLIENT = EveHttpxClient




import param
import panel as pn
from . import settings


class DefaultLayout(pn.GridBox):
    ncols = param.Integer(max(1, int(settings.GUI_WIDTH/200)))
    width = param.Integer(settings.GUI_WIDTH)
    
class EveModelBase(param.Parameterized):
    _panel = param.ClassSelector(pn.viewable.Viewable, default=None, precedence=-1)

    def make_panel(self):
        parameters = [k for k,v in self.params().items() if not k.startswith("_")]
        panel = pn.Param(self.param, width=settings.GUI_WIDTH,
                                   parameters=parameters,
                                   default_layout=pn.Card)
        return panel

    def panel(self):
        # return self.make_panel()
        if self._panel is None:
            self._panel = pn.panel(self.make_panel)
        return self._panel
    
    def _repr_mimebundle_(self, include=None, exclude=None):
        mimebundle = self.panel()._repr_mimebundle_(include, exclude)
        return mimebundle

    @property
    def gui(self):
        return self.panel()
    
    def show(self):
        return self.panel().show()

    def servable(self):
        return self.panel().servable()



import param
import panel as pn
from collections import defaultdict
from .eve_model import EveModelBase
from .resource import EveResource
from .http_client import DEFAULT_HTTP_CLIENT, EveHttpClient
from . import settings


class EveDomain(EveModelBase):
    name = param.String("EveDomain")
    _http_client = param.ClassSelector(EveHttpClient, precedence=-1)
    _selected = param.Selector(label="", default=None, allow_None=True)
    
    @classmethod
    def from_domain_def(cls, domain_def, domain_name="EveApp", http_client=None, sort_by_url=False):
        if http_client is None:
            http_client = DEFAULT_HTTP_CLIENT()
        if sort_by_url:
            domain_def = {v["url"]: v for v in domain_def.values()}
    
        sub_domains = defaultdict(dict)
        params = {}
        kwargs = {}
        for url, resource_def in sorted(domain_def.items(), key=lambda x: len(x[0])):
            sub_url, _, rest = url.partition("/")
            if rest:
                sub_domains[sub_url][rest] = resource_def
            else:
                resource = EveResource.from_resource_def(resource_def, url, http_client=http_client)
                params[sub_url] = param.ClassSelector(resource.__class__, default=resource, constant=True)
                kwargs[sub_url] = resource
        for url, domain_def in sub_domains.items():
            if url in params:
                for sub_url, resource_def in domain_def.items():
                    resource = EveResource.from_resource_def(resource_def, url, http_client=http_client)
                    kwargs[url+"_"+sub_url] = resource
                    params[url+"_"+sub_url] = param.ClassSelector(resource.__class__, default=resource, constant=True)
            else:
                sub_domain = EveDomain.from_domain_def(domain_def, url, http_client=http_client)
                kwargs[url] = sub_domain
                params[url] = param.ClassSelector(EveDomain, default=sub_domain, constant=True)
        
        klass = type(domain_name+"_domain", (cls,), params)
        instance = klass(name=domain_name, _http_client=http_client, **kwargs)
        instance.param._selected.objects = [""]+ list(params)
        instance._selected = ""
        return instance

    @param.depends("_selected")
    def sub_panel(self):
        if self._selected is None or self._selected=="":
            return pn.Column(height=200, width=int(settings.GUI_WIDTH))
        return getattr(self, self._selected).panel

    def sub_panel_view(self, name):
        def make_panel():
            panel = getattr(self, name).panel(show_client=False)
            return panel
        return make_panel

    def make_panel(self, show_client=True, tabs_location='above'):
        tabs = [(k.upper().replace("_", " "), getattr(self, k).make_panel(show_client=False, tabs_location="above")) for k,v in self.param.objects().items()
                 if isinstance(v, param.ClassSelector) and v.class_ in (EveDomain, EveResource)]
        tabs.sort(key=lambda x:len(x[0]))
        view = pn.Tabs(*tabs, dynamic=True, tabs_location=tabs_location)
        if show_client:
            view.append(("Settings", self._http_client.panel))
        
        return view

    def set_token(self, token):
        self._http_client.set_token(token)

    def __dir__(self):
        return list(self.params())




import param
import panel as pn
import pandas as pd
import json
import yaml
import itertools
from .field import SUPPORTED_SCHEMA_FIELDS, Validator
from io import StringIO, BytesIO

from .eve_model import EveModelBase
from .item import EveItem
from .http_client import DEFAULT_HTTP_CLIENT, EveHttpClient
from .page import EvePage, EvePageCache, PageZero
from . import settings

def read_csv(f):
    df = pd.read_csv(f).dropna(axis=1, how="all")
    return df.to_dict(orient="records")

file_readers = {
    "json": json.load,
    "yml": yaml.safe_load,
    "yaml": yaml.safe_load,
    "csv": read_csv,
}

def read_file(fname, f):
    name, _, ext = fname.rpartition(".")
    if ext in file_readers:
        data = file_readers[ext](f)
        if isinstance(data, list):
            return data
        elif isinstance(data, (dict,)):
            return [data]
    return []

class EveResource(EveModelBase):
    _http_client = param.ClassSelector(EveHttpClient, precedence=-1)
    _url = param.String(precedence=-1)
    _page_view_format = param.Selector(objects=["Table", "Widgets", "JSON"], default=settings.DEFAULT_VIEW_FORMAT,
                                         label="Display Format", precedence=1)
    _upload_errors = param.List(default=[])
    _resource = param.Dict(default={}, constant=True, precedence=-1)
    _schema = param.Dict(default={}, constant=True, precedence=-1)

    _cache = param.ClassSelector(class_=EvePageCache, default=EvePageCache())
    _item_class = param.ClassSelector(EveItem, is_instance=False, precedence=-1)
    _upload_buffer = param.List(default=[], precedence=-1)
    _file_buffer = param.ClassSelector(bytes)
    _filename = param.String()
    selection = param.ListSelector(default=[], objects=[], precedence=-1)
    
    filters = param.Dict(default={}, doc="Filters")
    columns = param.List(default=[], precedence=1)

    items_per_page = param.Integer(default=10, label="Items per page", precedence=1)
    _prev_page_button = param.Action(lambda self: self.decrement_page(), label="<<", precedence=1)
    page_number = param.Integer(default=0, bounds=(0, None), label="", doc="Page number", precedence=2)
    _next_page_button = param.Action(lambda self: self.increment_page(), label=">>", precedence=3)
    
    @classmethod
    def from_resource_def(cls, resource_def, resource_name, http_client=None, items=[]):
        resource = dict(resource_def)
        schema = resource.pop("schema")
        if http_client is None:
            http_client = DEFAULT_HTTP_CLIENT()
        item = EveItem.from_schema(resource["item_title"], schema, resource["url"] , http_client=http_client)
        item_class = item.__class__
        params = dict(
            name = resource["resource_title"],
            _url = resource["url"],
            _http_client = http_client,
            _item_class = item_class,
            _resource = resource,
            _schema = schema,
            columns=list(schema)
        )
        return cls(**params)

    def __getitem__(self, key):
        if key in self._cache:
            return self._cache[key]
        data = self._http_client.get("/".join([self._url, key]))
        if data:
            item = self.make_item(**data)
            return item
        raise KeyError
    
    def __setitem__(self, key, value):
        self._cache[key] = value

    def make_item(self, **kwargs):
        return self._item_class(**kwargs, _http_client=self._http_client)

    @param.depends("_upload_buffer")
    def upload_view(self):
        clear_button = pn.widgets.Button(name="Clear buffer", 
                                        button_type="warning", 
                                        width=int(settings.GUI_WIDTH/4))
        clear_button.on_click(lambda event: self.clear_buffer())

        upload_file = pn.widgets.FileInput(accept=",".join([f".{ext}" for ext in file_readers]), 
                                            width=int(settings.GUI_WIDTH/4))
        upload_file.link(self, filename="_filename", value="_file_buffer")
        upload_file_button = pn.widgets.Button(name="Read file", button_type="primary", width=int(settings.GUI_WIDTH/4))
        upload_file_button.on_click(lambda event: self._read_file_buffer())

        upload_preview = pn.pane.JSON(self._upload_buffer, name='Upload Buffer',
                                    height=int(settings.GUI_HEIGHT/2),
                                    width=int(settings.GUI_WIDTH), theme="light")
        upload_button = pn.widgets.Button(name="Insert to DB", button_type="success", 
                                            width=int(settings.GUI_WIDTH/4))
        upload_button.on_click(lambda event: self.flush_buffer())
        read_clipboard_button = pn.widgets.Button(name="Read Clipboard", button_type="primary",
                                                     width=int(settings.GUI_WIDTH/4))
        read_clipboard_button.on_click(lambda event: self.read_clipboard())

        first_row_buttons = pn.Row(upload_file,read_clipboard_button)
        second_row_buttons = pn.Row(clear_button, upload_button)
        input_buttons = pn.Column(first_row_buttons, second_row_buttons)
   
        upload_view = pn.Column(input_buttons, self.upload_errors, "### Upload buffer" ,upload_preview)
        return upload_view

    @param.depends("page_number", "_cache", "_page_view_format")
    def current_page_view(self):
        page = self.get_page(self.page_number)
        if page is None:
            return pn.panel(f"## No data for page {self.page_number}.")
        return getattr(page, self._page_view_format.lower()+"_view", page.panel)()
    
    @param.depends("_upload_errors")
    def upload_errors(self):
        alerts = [pn.pane.Alert(err, alert_type="danger") for err in self._upload_errors]
        return pn.Column(*alerts, height=50, width=int(settings.GUI_WIDTH-30))

    def make_panel(self, show_client=True, tabs_location='above'):
        buttons = pn.Param(self.param, parameters=["_prev_page_button", "page_number", "_next_page_button"],
                             default_layout=pn.Row, name="", width=int(settings.GUI_WIDTH/3))

        column_select = pn.Param(self.param.columns, width=int(settings.GUI_WIDTH-30), 
                                 widgets={"columns": {"type": pn.widgets.MultiChoice,
                                                    "options": list(self._schema)+settings.META_COLUMNS,
                                                    "width": int(settings.GUI_WIDTH-30)}})

        page_settings = pn.Column(pn.Row(self.param.items_per_page, self.param.filters, self.param._page_view_format,
                                             width=int(settings.GUI_WIDTH-50)),
                                     column_select,
                                      width=int(settings.GUI_WIDTH-10))
        if show_client:
            page_settings.append("### HTTP client")
            page_settings.append(pn.layout.Divider())
            page_settings.append(self._http_client.panel)
        
        header = pn.Row(f"## {self.name} resource",
            pn.Spacer(sizing_mode='stretch_both'),
            buttons,
            pn.Spacer(sizing_mode='stretch_both'),
            self._http_client.busy_indicator,)

        view = pn.Column(
            header,
            self._http_client.messages,
            pn.Tabs(("Data", self.current_page_view), 
                    ("Settings", page_settings),
                    ("Upload", self.upload_view),
                    dynamic=True, tabs_location=tabs_location,
                    width=int(settings.GUI_WIDTH),
                         )
        )
        
        return view
    
    @property
    def projection(self):
        return { k:1 for k in self.columns if k not in settings.META_COLUMNS}

    def read_clipboard(self):
        from pandas.io.clipboard import clipboard_get
        try:
            self._upload_buffer = self.filter_docs(json.loads(clipboard_get()))
        except Exception as e:
            print(e)

    def read_file(self, fname, f=None):
        if f is None:
            f = open(f, "rb")
        data = read_file(fname, f)
        self._upload_buffer = self.filter_docs(data)
        return data

    @param.depends("_file_buffer", watch=True)
    def _read_file_buffer(self):
        sio = BytesIO(self._file_buffer)
        self.read_file(self._filename, sio)

    def filter_docs(self, docs):
        if isinstance(docs, dict):
            docs = [docs]
        return [{k:v for k,v in doc.items() if k in self._schema} for doc in docs]

    @property
    def gui(self):
        return self.panel()
    
    @property
    def df(self):
        return self.to_dataframe()

    def keys(self):
        for i in itertools.count():
            idx = i+1
            items = self.find(projection={"_id":1}, max_results=5000, page_number=idx)
            if not items:
                break
            for item in items:
                yield item._id

    def values(self):
        for i in itertools.count():
            idx = i+1
            page = self.get_page(idx)
            if not len(page):
                break
            yield from page.values()

    def items(self):
        for i in itertools.count():
            idx = i+1
            page = self.get_page(idx)
            if not len(page):
                break
            yield from page.items()

    def new_item(self, data={}):
        item = self.item_class(**data)
        self[item._id] = item

    def to_records(self):
        return [item.to_dict() for item in self.values()]
    
    def to_dataframe(self):
        df = pd.concat([page.to_dataframe() for page in self.values()])
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df
    
    def pull(self):
        for i in itertools.count():
            idx = i+1
            if not self.pull_page(idx):
                break
            
    def push(self, idxs=None):
        if idxs is None:
            idxs = list(self._cache.keys())
        for idx in idxs:
            self._cache[idx].push()

    def find(self, query={}, projection={}, max_results=25, page_number=1):
        resp = self._http_client.get(self._url, where=json.dumps(query),
                 projection=json.dumps(projection), max_results=max_results, page=page_number)
        docs = []
        if resp and "_items" in resp:
            docs = resp["_items"]
        return docs
        

    def find_page(self, query={}, projection={}, max_results=25, page_number=1):
        docs = self.find(query=query, projection=projection, max_results=max_results, page_number=page_number)
        items = [self.make_item(**doc) for doc in docs]
        page = EvePage(name=f'{self._url.replace("/", ".")} page {page_number}',
                                    _items={item._id:item for item in items},
                                    _columns=self.columns)
        return page

    def find_df(self, query={}, projection={}, max_results=25, page_number=1):
        page = self.find_page(query=query, projection=projection, max_results=max_results, page_number=page_number)
        df = page.to_dataframe()
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df

    def find_one(self, query={}, projection={}):
        docs = self.find(query=query, projection=projection)
        if docs:
            return docs[0]

    def find_one_item(self, query={}, projection={}):
        doc = self.find_one(query=query, projection=projection)
        if doc:
            return self.make_item(**doc)

    def post(self, docs):
        return self._cleint.post(self._url, json=docs)

    # def add_data(self, data):
    #     if isinstance(data, dict):
    #         data = [data]
    #     docs = []
    #     for doc in data:
    #         if isinstance(doc, dict):
    #             docs.append(doc)
    #         elif isinstance(doc, EveItem):
    #             docs.append(doc.to_dict())
    #     valid = _validate_docs(docs)
    #     self._upload_buffer = self._upload_buffer + docs

    def _validate_docs(self, docs):
        schema = {name: {k:v for k,v in field.items() if k in SUPPORTED_SCHEMA_FIELDS} for name, field in self._schema.items()}
        v = Validator(schema)
        valid = []
        rejected = []
        errors = []
        for doc in docs:
            if v.validate(doc):
                valid.append(doc)
            else:
                rejected.append(doc)
                errors.append(v.errors)
        return valid, rejected, errors

    def clear_buffer(self):
        self._upload_buffer = []

    def flush_buffer(self):
        valid, rejected, errors = self._validate_docs(self._upload_buffer)
        if valid:
            self._http_client.post(self._url, data=json.dumps(valid))
        self._upload_buffer = rejected
        self._upload_errors = [str(err) for err in errors]
        return valid

    def pull_page(self, idx=0):
        if not idx:
            self._cache[idx] = PageZero()
            return False
        page = self.find_page(query=self.filters, projection=self.projection, 
                            max_results=self.items_per_page, page_number=idx)
        if page._items:
            self._cache[idx] = page
            return True
        return False

    def push_page(self, idx):
        if not idx in self._cache or len(self._cache[idx]):
            return
        self._cache[idx].push()

    def get_page(self, idx):
        if idx not in self._cache or not len(self._cache[idx]):
            self.pull_page(idx)
        return self._cache.get(idx, EvePage(name="Place holder", _columns=self.columns))

    def get_page_records(self, idx):
        return self.get_page(idx).to_records()

    def get_page_df(self, idx):
        df = self.get_page(idx).to_dataframe()
        df = df[[col for col in self.columns if col in df.columns]]
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df

    def increment_page(self):
        self.page_number = self.page_number + 1

    def next_page(self):
        self.increment_page()
        return self.current_page()

    def current_page(self):
        return self.get_page(self.page_number)

    def decrement_page(self):
        if self.page_number > 1:
            try:
                self.page_number = self.page_number - 1
            except:
                pass
        
    def previous_page(self):
        self.decrement_page()
        return self.current_page()

    @param.depends("items_per_page", "filters", "columns", watch=True)
    def _query_settings_changed(self):
        self._cache = EvePageCache()



import param
import panel as pn
import pandas as pd

from io import StringIO, BytesIO

from .eve_model import EveModelBase
from .item import EveItem
from .http_client import DEFAULT_HTTP_CLIENT, EveHttpClient
from . import settings


class EvePage(EveModelBase):
    _columns = param.List(default=["_id"])
    _items = param.Dict(default={})

    def __getitem__(self, key):
        return self._items[key]

    def __setitem__(self, key, value):
        self._items[key] = value

    def __contains__(self, key):
        return key in self._items
    
    def __len__(self):
        return len(self._items)
    
    def __bool__(self):
        return bool(len(self))
    
    @property
    def df(self):
        return self.to_dataframe()
    
    def keys(self):
        yield from self._items.keys()
    
    def values(self):
        yield from self._items.values()

    def items(self):
        yield from self._items.items()
               
    def to_records(self):
        return [item.to_dict() for item in self.values()]
    
    def to_dataframe(self):
        df = pd.DataFrame(self.to_records(), columns=self._columns)
        if "_id" in df.columns:
            df = df.set_index("_id")
        return df
    
    def push(self, names=None):
        if names is None:
            names = self.keys()
        for name in names:
            self._items[name].push()
    
    def pull(self, names=None, version=1):
        if names is None:
            names = self.keys()
        for name in names:
            self._items[name].pull(version=version)

    @param.depends("_items")
    def widgets_view(self):
        if not len(self._items):
            return pn.Column("## No items to display.")
        # n = settings.WIDGET_VIEW_ITEMS
        items = [(item.name, item.panel()) for item in self._items.values()]
        # if len(items)<n:
        #     return pn.GridBox(*items,
        #                 height=int(settings.GUI_HEIGHT-10),
        #                 width=settings.GUI_WIDTH, ncols=max(1, int(settings.GUI_WIDTH/300)))

        # sub_pages = [items[i*n:(i+1)*n] for i in range(len(items)//n+1)]
        # tabs = [(str(i),pn.GridBox(*sub_page,
        #                 height=int(settings.GUI_HEIGHT-10),
        #                 width=settings.GUI_WIDTH, ncols=max(1, int(settings.GUI_WIDTH/300)))) for i,sub_page in enumerate(sub_pages)]
        # cards = [pn.Accordion(*sub_page,toggle=True,
        #                 height=int(settings.GUI_HEIGHT-10),) for sub_page in sub_pages]
        # view = pn.Tabs(*tabs, dynamic=True)
        # view = pn.GridBox(*cards, ncols=max(1, int(settings.GUI_WIDTH/300)))
        # view = pn.Accordion(*items, toggle=True, height=int(settings.GUI_HEIGHT-10) )
        view = pn.Tabs(*items, dynamic=True, height=int(settings.GUI_HEIGHT-10), width=settings.GUI_WIDTH)
        return view

    @param.depends("_items")
    def table_view(self):
        if not len(self._items):
            return pn.Column("## No items to display.")
        df = self.to_dataframe()
        return pn.widgets.DataFrame(df, disabled=True, width=settings.GUI_WIDTH,
                                        height=int(settings.GUI_HEIGHT-30))

    @param.depends("_items")
    def json_view(self):
        return pn.pane.JSON(self.to_records(), theme="light",
            width=settings.GUI_WIDTH, height=int(settings.GUI_HEIGHT-30))

    def panel(self):
        tabs = pn.Tabs(("Table", self.table_view), 
                        ("Widgets", self.widgets_view), 
                        ("JSON", self.json_view),
                        height=int(settings.GUI_HEIGHT),
                        width=settings.GUI_WIDTH,
                        dynamic=True)
        return pn.Column(f"## {self.name}", tabs)


class PageZero(EvePage):
    @param.depends("_items")
    def widgets_view(self):
        return self.panel()

    @param.depends("_items")
    def table_view(self):
        return self.panel()

    @param.depends("_items")
    def json_view(self):
        return self.panel()

    def panel(self):
        return pn.Column(
                pn.layout.Divider(width=settings.GUI_WIDTH),
                "### You are on the landing page for this resource, no data here.",
                "TIP 1: Use the >> button to load the first page.",
                "TIP 2: You can use the settings tab to change what data is loaded and how it is displayed.",
                "TIP 3: If you just want to upload data, you can go directly to the upload tab. ",
                pn.layout.Divider(width=settings.GUI_WIDTH),
                width=settings.GUI_WIDTH,
                height=300,
                )

class EvePageCache(param.Parameterized):
    _pages = param.Dict(default={})

    def __getitem__(self, key):
        if isinstance(key, str):
            for page in self._pages.values():
                if key in page:
                    return page[key]
            else:
                raise KeyError
        elif isinstance(key, int):
            return self._pages[key]

    def __setitem__(self, key, value):
        self._pages[key] = value

    def __contains__(self, key):
        if isinstance(key, str):
            for page in self._pages.values():
                if key in page:
                    return True
            else:
                return False
        return key in self._pages

    def get(self, key, fallback=None):
        return self._pages.get(key, fallback)
        # if key in self._pages:
        #     return self._pages[key]
        # else:
        #     return fallback

    def keys(self):
        yield from self._pages.keys()

    def values(self):
        yield from self._pages.values()

    def items(self):
        yield from self._pages.items()




import param
import panel as pn
from bson import ObjectId
from .eve_model import EveModelBase, DefaultLayout
from .types import TYPE_MAPPING
from .widgets import get_widget
from .http_client import DEFAULT_HTTP_CLIENT, EveHttpClient
from .field import EveField
from . import settings


class EveItem(EveModelBase):
    _http_client = param.ClassSelector(EveHttpClient, precedence=-1)
    _resource_url = param.String(precedence=-1)
    _etag = param.String( allow_None=True,precedence=15)
    _version = param.Integer(default=1, bounds=(1,None), allow_None=True, precedence=13)
    _latest_version = param.Integer(default=1, bounds=(1,None), allow_None=True, constant=True, precedence=14)

    _save = param.Action(lambda self: self.push(), label="Save", precedence=16)
    _delete = param.Action(lambda self: self.delete(), label="Delete", precedence=16)
    _clone = param.Action(lambda self: self.clone(), label="Clone", precedence=16)

    def __init__(self, **params):
        params["_id"] = params.get("_id", str(ObjectId()))
        if "name" not in params:
            params["name"] = f'{self.__class__.__name__}_{params["_id"]}'
        params = {k:v for k,v in params.items() if hasattr(self, k)}
        super().__init__(**params)
  
    @classmethod
    def from_schema(cls, name, schema, resource_url, http_client=None, data={}):
        if http_client is None:
            http_client = DEFAULT_HTTP_CLIENT()
        params = dict(
            _schema=param.Dict(default=schema, allow_None=False, constant=True, precedence=-1),
            
            _resource_url=param.String(default=resource_url,precedence=-1),
        )
        _widgets ={"_etag":{"type": pn.widgets.TextInput, "disabled": True},
                    "_version": {"type": pn.widgets.IntInput, "disabled": False}}

        for field_name, field_schema in schema.items():
            kwargs = {"precedence": 10}
            if field_name=="_id":
                kwargs["precedence"] = 1
            extended_name = f"{name.title()}{field_name.title()}"

            if "allowed" in field_schema:
                class_ = param.Selector
                kwargs["objects"] = field_schema["allowed"]
   
            elif field_schema["type"] in TYPE_MAPPING:
                class_ = EveField(extended_name, field_schema, TYPE_MAPPING[field_schema["type"]])
            else:
                continue
            if "default" in field_schema:
                kwargs["default"] = field_schema["default"]

            widget = get_widget(extended_name, field_schema)
            if widget is not None:
                _widgets[field_name] = widget

            kwargs["allow_None"] = field_schema.get("nullable", False) or not field_schema.get("required", False)
            
            bounds = (field_schema.get("min", None), field_schema.get("max", None))
            if any(bounds):
                kwargs["bounds"] = bounds
            kwargs["readonly"] = field_schema.get("readonly", False)
            params[field_name] = class_(**kwargs)
        params["_widgets"] = param.Dict(default=_widgets, constant=True, precedence=-1)
        klass = type(name, (EveItem,), params)
        
        return klass(_schema=schema, _widgets=_widgets,_resource_url=resource_url, _http_client=http_client, **data)
    
    def make_panel(self):
        header = pn.Column(pn.layout.Divider(), f"### {self.name}",)
        buttons = pn.Param(self.param, parameters=["_save", "_delete"], 
                            widgets={"_delete": {"type": pn.widgets.Button, "button_type":"danger"},
                            "_save": {"type": pn.widgets.Button, "button_type":"success"},
                            },
                            show_name=False, default_layout=pn.Row)
        
        editors = pn.Param(self.param, show_name=False, default_layout=DefaultLayout, width=settings.GUI_WIDTH, 
                        widgets=self._widgets, parameters=list(self._schema)+settings.META_COLUMNS)
        return pn.Column(header, editors, buttons)
    
    def save(self):
        self.push()

    def to_record(self):
        return {k: getattr(self, k) for k in self._schema}

    def to_dict(self):
        return self.to_record()

    @param.depends("_version", watch=True)
    def pull(self):
        if self._version is None:
            return
        data = self._http_client.get("/".join([self._resource_url, self._id]), version=self._version)
        if not data:
            return
        for k,v in data.items():
            if hasattr(self, k):
                try:
                    param = getattr(self.param, k)
                    if param.constant or param.readonly:
                        setattr(self, getattr(param, "_internal_name"), v)
                    else:
                        setattr(self, k, v) 
                except Exception as e:
                    print(e)
                    pass

    def version(self, version):
        self._version = version
        # self.pull()
        return self

    def versions(self):
        data = self._http_client.get("/".join([self._resource_url, self._id]), version='all')
        if not data:
            return []
        return [self.__class__(**doc, _http_client=self._http_client, _resource_url=self._resource_url)
                     for doc in data.get("_items", [])]

    def version_diffs(self):
        data = self._http_client.get("/".join([self._resource_url, self._id]), version='diffs')
        if not data:
            return []
        return [self.__class__(**doc, _http_client=self._http_client, _resource_url=self._resource_url)
                     for doc in data.get("_items", [])]

    def push(self):
        url = "/".join([self._resource_url, self._id])
        data = {"_id": self._id,}
        if self._version==self._latest_version:
            data["_etag"] = self._etag

        for k in self._schema:
            data[k] = getattr(self, k)
        self._http_client.put(url, data)
        self.pull()

    def patch(self, fields):
        url = "/".join([self._resource_url, self._id])
        data = {"_id": self._id, "_etag": self._etag}
        for k in fields:
            data[k] = getattr(self, k)
        self._http_client.patch(url, data)
    
    def clone(self):
        data = {k: getattr(self, k) for k in self._schema}
        return self.__class__(**data)

    def delete(self):
        url = "/".join([self._resource_url, self._id])
        data = {"_id": self._id, "_etag": self._etag}
        return self._http_client.delete(url, data)

    def __repr__(self):
        return f"{self.__class__.__name__}(_id={self._id or self.name})"
    


import param
import panel as pn
import base64
import pkg_resources
from .eve_model import EveModelBase

class EveAuthBase(param.Parameterized):

    def get_headers(self):
        return {}

    def login(self):
        return True

    def panel(self):
        return pn.Column()
    
    def set_token(self, token):
        self.token = token


class EveBasicAuth(EveAuthBase):
    username = param.String(precedence=1)
    password = param.String(precedence=2)
    token = param.ClassSelector(bytes, default=b"")

    def login(self):
        return True

    def get_headers(self):
        token = base64.b64encode(f"{self.username}:{self.password}".encode())
        return {"Authorization": f"Basic {token}"}

    def panel(self):
        return pn.Param(self.param, parameters=["username","password"], widgets={"password": pn.widgets.PasswordInput})

class EveBearerAuth(EveAuthBase):
    token = param.String()

    def get_headers(self):
        if self.token:
            return {"Authorization": f"Bearer {self.token}"}
        else:
            return {}

    def login(self):
        return bool(self.token)

    def panel(self):
        return pn.panel(self.param.token)


AUTH_CLASSES = {
    "Basic": EveBasicAuth,
    "Bearer": EveBearerAuth, 
}

class AuthSelector(EveAuthBase):
    auth_options = { k:v() for k,v in AUTH_CLASSES.items() } #[v() for v in AUTH_CLASSES.values()]

    _auth_object = param.ObjectSelector(default=auth_options[list(auth_options)[0]], objects=auth_options, label="Authentication")
    # _auth = param.ClassSelector(EveAuthBase, default=EveBasicAuth())
    @property
    def token(self):
        return getattr(self._auth_object, "token", "")

    @token.setter
    def token(self, val):
        setattr(self._auth_object, "token", val)

    def login(self):
        return self._auth_object.login()

    def get_headers(self):
        return self._auth_object.get_headers()
    
    @param.depends("_auth_object")
    def auth_view(self):
        if self._auth_object is None or not hasattr(self._auth_object, "panel"):
            return pn.Column()

        return self._auth_object.panel()

    def set_auth_by_name(self, name):
        self._auth_object = self.param._auth_object.names[name]

    def panel(self):
        return pn.Column(self.param._auth_object, self.auth_view)
    
    def __init__(self, **params):
        super().__init__( **params)
        for entry_point in pkg_resources.iter_entry_points('eve_panel.auth'):
            AUTH_CLASSES[entry_point.name] = entry_point.load()
        auth_options = { k:v() for k,v in AUTH_CLASSES.items() }
        self.param._auth_object.objects = list(auth_options.values())
        self.param._auth_object.names = auth_options
        self._auth_object = auth_options[list(auth_options)[0]]



import param
from eve.io.mongo.validation import Validator

SUPPORTED_SCHEMA_FIELDS = ["type", "schema", "required", "readonly", "valueschema", "keyschema", "regex", 
"minlength", "maxlength", "min", "max", "allowed","items", "empty", "nullable",
]

def EveField(name, schema, klass):
    if isinstance(klass, param.ClassSelector):
        return klass
    if not isinstance(klass, type):
        return klass
    schema = {k:v for k,v in schema.items() if k in SUPPORTED_SCHEMA_FIELDS}
    validator = Validator({"value": schema})

    def _validate(self, val):
        if self.allow_None and val is None:
            return
        if self.owner is None:
            return
        if self.name is None:
            return
        if not validator.validate({"value": val}):
            sep = "\n"
            errors = [f"Cannot set \'{self.owner.name}.{self.name}\' to \'{val}\' of type {type(val)}."]
            for k,v in validator.errors.items():
                errors.append(f"{k} {v}")
            if len(errors)<=2:
                sep = ". "
            raise ValueError(" ".join(errors))
            
    params = {
        "_validate": _validate,
        
    }

    return type(f"Eve{name.title()}{klass.__name__}Field", (klass,), params)
    


"""Top-level package for Eve-Panel."""
from .eve_client import EveClient
from .domain import EveDomain
from .resource import EveResource
from .item import EveItem

from .utils import from_app_config

import panel as pn

def notebook():
    return pn.extension()

__author__ = """Yossi Mosbacher"""
__email__ = 'joe.mosbacher@gmail.com'
__version__ = '0.1.0'




import param
import panel as pn
import eve
from .http_client import EveHttpClient, DEFAULT_HTTP_CLIENT
from .eve_model import EveModelBase
from .domain import EveDomain
from pprint import pprint

class EveClient(EveModelBase):
    _http_client = param.ClassSelector(EveHttpClient, default=DEFAULT_HTTP_CLIENT())
    domain = param.ClassSelector(EveDomain)

    @classmethod
    def from_app(cls, app, name="EveDomain", sort_by_url=False):
        settings = app.config
        # pprint(settings["DOMAIN"])
        http_client = DEFAULT_HTTP_CLIENT.from_app_settings(settings)
        domain=EveDomain.from_domain_def(domain_def=settings["DOMAIN"], domain_name=name, http_client=http_client, sort_by_url=sort_by_url)
        return cls(domain=domain, _http_client=http_client)

    @classmethod
    def from_app_settings(cls, settings, sort_by_url=False):
        app = eve.Eve(settings)
        return cls.from_app(app, sort_by_url=sort_by_url)
        
    @property
    def db(self):
        return self.domain

    def make_panel(self, show_client=False):
        return self.domain.make_panel(show_client=show_client)

    def set_token(self, token):
        self._http_client.set_token(token)


import param
from bson import objectid

class CoerceClassSelector(param.ClassSelector):
    def __set__(self, obj, val):
        try:
            val = self.class_(val)
        except:
            pass
        super().__set__(obj, val)
        
def objectid_param(**kwargs):
    return CoerceClassSelector(str, constant=True, **kwargs)

def bytes_param(**kwargs):
    return param.ClassSelector(bytes, **kwargs)

def set_param(**kwargs):
    return param.ClassSelector(set, **kwargs)

TYPE_MAPPING = {
    "objectid": objectid_param,
    "boolean": param.Boolean,
    "binary": bytes_param,
    "date": param.Date,
    "datetime": param.Date,
    "dict": param.Dict,
    "float": param.Number,
    "integer": param.Integer,
    "list": param.List,
    "number": param.Number,
    "set": set_param,
    "string": param.String,
    "media": bytes_param,
}



import param
import panel as pn
from panel.widgets import LiteralInput
import json
import ast
from eve.io.mongo.validation import Validator

class LiteralSchemaInputBase(LiteralInput):
    
    def validate_schema(self, value):
        return True
    
    def _process_property_change(self, msg):
        msg = super(LiteralSchemaInputBase, self)._process_property_change(msg)
        if msg['value'] == self.value:
            return msg
        new_state = ''
        if 'value' in msg:
            value = msg.pop('value')    
            if not self.validate_schema(value):
                new_state = ' (invalid)'
                value = self.value
            msg['value'] = value
            msg['name'] = msg.get('title', self.name).replace(self._state, '').replace(new_state, '') + new_state
            self._state = new_state
            self.param.trigger('name')
        return msg

def LiteralSchemaInput(name, schema, type_=None):
    validator = Validator({"value": schema})
    def validate_schema(self, value):
        return validator.validate({"value": value})
    params = {
        "validate_schema": validate_schema,
        "type": type_,
    }
    return type(name+"InputWidget", (LiteralSchemaInputBase,), params)

WIDGET_MAPPING = {
    "media": pn.widgets.FileInput,
}

def get_widget(name, schema):
    if schema["type"]=="dict" and "schema" in schema:
        return LiteralSchemaInput(name, schema, dict)
    elif schema["type"]=="list" and "schema" in schema:
        return LiteralSchemaInput(name, schema, list)
    else:
        return WIDGET_MAPPING.get(schema["type"], None)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eve_panel', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['eve>=1.1.3,<2.0.0',
 'httpx>=0.16.1,<0.17.0',
 'pandas>=1.1.3,<2.0.0',
 'panel>=0.9.7,<0.10.0']

setup_kwargs = {
    'name': 'eve-panel',
    'version': '0.3.0',
    'description': 'Top-level package for Eve-Panel.',
    'long_description': '=========\nEve-Panel\n=========\n\n\n.. image:: https://img.shields.io/pypi/v/eve_panel.svg\n        :target: https://pypi.python.org/pypi/eve_panel\n\n.. image:: https://img.shields.io/travis/jmosbacher/eve_panel.svg\n        :target: https://travis-ci.com/jmosbacher/eve_panel\n\n.. image:: https://readthedocs.org/projects/eve-panel/badge/?version=latest\n        :target: https://eve-panel.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\nDynamically create an httpx based client for any Eve api that uses Param for type enforcement and Panel for GUIs.\nThis is just a prototype package,features will slowly be added as i need them for my own purposes.\nThe api is expected to change without warning based on my needs but feel free to fork or copy parts and adapt to your own needs.\n\nTo view the widgets in a notebook you will need to install the pyviz plugin. For details, see panel docs.\n.. code-block:: python\n\n        import eve\n        from eve_panel import EveApiClient, EveDomain\n\n        app = eve.Eve()\n\n        client = eve_panel.EveApiClient.from_app_config(app.config, address="http://localhost:5000")\n        client.auth.token = "my-secret-token"\n\n        api = eve_panel.EveDomain.from_domain_def("my_api_name", app.config["DOMAIN"], client=client)\n\n        # show a resources gui\n        api.resource_name \n\n        # get a specific item\n        api.resource_name["item_id"]\n\n        # get current page data\n        api.resource_name.current_page()\n\n        # get next page data\n        api.resource_name.next_page()\n\n        # get previous page data\n        api.resource_name.previous_page()\n\n\n* Free software: MIT\n* Documentation: https://eve-panel.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n',
    'author': 'Yossi Mosbacher',
    'author_email': 'joe.mosbacher@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jmosbacher/eve_panel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
